%%%-------------------------------------------------------------------
%%% File    : connection_pool.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : Connection pool
%%%
%%% Created :  9 Dec 2002 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id$
%%
%% $Log$
%%
%%**
%% 
%% @doc It is a connection pool
%% <p>Now it supports two modes
%% <dl>
%%    <dt><code>shared</code></dt>
%%       <dd>client requests spread between connection servers. On request
%%           server with minimal number of users selected.
%%       </dd>
%%    <dt><code>exclusive</code></dt>
%%       <dd>only one client per connection server. Servers are selected in
%%           round-robin manner.
%%       </dd>
%% </dl>
%% </p>
%%  <p> Client can return server explicitly with
%%      <code>disconnect/1</code> or will be returned implicitly after its
%%      exit.
%%  </p>
%% @end
%%*
-module(connection_pool).

-behaviour(gen_server).
%%--------------------------------------------------------------------
%% Include files
%%--------------------------------------------------------------------

%%--------------------------------------------------------------------
%% External exports
-export([start_link/2, stop/1, stop_insure/1, info/1, connect/1, connect/2, disconnect/1, call/2, call/3, cast/2, conn/2]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2, code_change/3]).

-export([worker_wrapper/3]).

-import(svglib.getopt).
-import(svglib.misc).
-import(svglib.dbg).

-record(state, {pool,% {Workers, Users} Users = [{User, Worker}]
		pool_type = exclusive, % exclusive | shared
		user_queue = queue:new(), %waiting for worker
		pool_sup,% pool superviser
		pool_start_len=0,
		pool_max_len=1,
		worker_spec,
		num_started=0 %started workers but not yet initialized
	       }).

-record(pool, {workers,
	       users=[],
	       num_workers=0
	      }).

-record(conn_id, {pool_pid,
		  worker_pid
		 }).

connect(Pool) ->
  connect(Pool, infinity).

%%**
%% @spec connect( Pool, Timeout::integer() ) -> connection() | error()
%%
%%       Pool    = name() | pid()
%%       
%% @doc Request connection from connection pool
%%
%% @end
%%*
connect(Pool, Timeout) ->
  Ref = make_ref(),
  case catch gen_server:call(Pool, {connect, self(), Ref}, Timeout) of
    {'EXIT', {timeout, _}} ->
      %% it must remove us from queue
      gen_server:cast(Pool, {connect_timeout, self(), Ref}),
      exit(timeout);
    {'EXIT', Reson} ->
      exit(Reson);
    Result ->
      Result
  end.



%%**
%% @spec disconnect( Connection::connection() ) -> ok
%%
%% @doc Return server to connection pool
%%
%% @end
%%*
disconnect(Cid=#conn_id{pool_pid=Pool, worker_pid=Worker}) ->
  gen_server:cast(Pool, {disconnect, self(), Worker}).


%%**
%% @spec conn( worker, Connection::connection() ) -> pid()
%%
%% @doc Get server pid from connection returned by <code>connect</code>
%%
%% @end
%%*
conn(worker, C=#conn_id{worker_pid=W}) ->
  W.

info(Pool) ->
  gen_server:call(Pool, {info, state}).

%% gen_server bridge
call(Cid=#conn_id{worker_pid=Worker}, Request) ->
  gen_server:call(Worker, Request).

call(Cid=#conn_id{worker_pid=Worker}, Request, Timeout) ->
  gen_server:call(Worker, Request, Timeout).

cast(Cid=#conn_id{worker_pid=Worker}, Request) ->
  gen_server:cast(Worker, Request).

%%====================================================================
%% External functions
%%====================================================================


%%**
%% @spec start_link( Name, Opts ) -> {ok, Pid} | error()
%%
%%       Name = {local, atom()} | {global, atom()}
%%       Opts = [{{worker_spec, WorkerSpec}| OptionalOpts]
%%       OptionalOpts = {supervisor_spec, SuperSPec}
%%                       | {pool_type, PoolType}
%%                       | {pool_max_len, integer()}
%%                       | {pool_start_len, integer()}
%%       
%% @doc Start connection pool.
%%      <h4>Defaults</h4>
%%      <dl>
%%      <dt><code>supervisor_spec</code></dt>
%%         <dd><code>{simple_one_for_one, 1,  1000}</code></dd>
%%      <dt><code>pool_type</code></dt>
%%         <dd>exclusive</dd>
%%      <dt><code>pool_start_len</code></dt>
%%         <dd>0</dd>
%%      <dt><code>pool_max_len</code></dt>
%%         <dd>1</dd>
%%      </dl>
%% @end
%%*

start_link(Name, Opts) ->
  gen_server:start_link(Name, ?MODULE, Opts, []).

stop_insure(Srv) ->
  Ref = erlang:monitor(process, Srv),
  stop(Srv),
  receive
    {'DOWN', Ref, _, _, _} ->
      ok
  end.

stop(Srv) ->
  Srv ! {self(), stop},
  ok.

%%====================================================================
%% Server functions
%%====================================================================

%%--------------------------------------------------------------------
%% Function: init/1
%% Description: Initiates the server
%% Returns: {ok, State}          |
%%          {ok, State, Timeout} |
%%          ignore               |
%%          {stop, Reason}
%%--------------------------------------------------------------------
init(UserOpts) ->
  process_flag(trap_exit, true),
  case catch
    begin
      Opts = options(UserOpts),
      [SupSpec, WorkerSpec, Max, Start, PoolType] =
	[getopt:value(Key, Opts) 
	 || Key <- [supervisor_spec, worker_spec, pool_max_len, pool_start_len, pool_type]],
      {ok, PoolSup} = start_pool_sup(SupSpec, WorkerSpec),
      Pool = pool_new(PoolType),
      init_pool(PoolSup, Start),
      {ok, #state{pool=Pool, pool_type=PoolType, pool_sup=PoolSup,
		  pool_max_len=Max, pool_start_len=Start,
		  worker_spec=WorkerSpec, num_started=Start}}
    end of
    Result = {ok, _} ->
      Result;
    Error ->
      {stop, misc:better_error(Error)}
  end.

%%--------------------------------------------------------------------
%% Function: handle_call/3
%% Description: Handling call messages
%% Returns: {reply, Reply, State}          |
%%          {reply, Reply, State, Timeout} |
%%          {noreply, State}               |
%%          {noreply, State, Timeout}      |
%%          {stop, Reason, Reply, State}   | (terminate/2 is called)
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_call({connect, User, Ref}, From,
	    S=#state{pool=P, pool_sup=PS, pool_max_len=Max,
		     user_queue=UQ, num_started=NumS}) ->
    case
      case {pool_has_free_workers(P),
	    ((pool_num_workers(P) + NumS) < Max)} of
	{true, _} ->
	  {pool_select_worker(P), NumS};
	{false, true} ->
	  start_worker(PS),
	  {{to_queue, P}, NumS+1};
	{false, false} -> %max limit of workers exceeded
	  {pool_select_worker(P), NumS}
      end of
      {{{ok, Worker}, P1}, _} ->
	NewP = pool_add_user(Worker, User, Ref, P1),
	{reply, {ok, conn_id(Worker, self())}, S#state{pool=NewP}};
      {{to_queue, _}, NewNumS} ->
	NewUQ = user_queue_in({User, From, Ref}, UQ),
	{noreply, S#state{user_queue=NewUQ, num_started=NewNumS}};
      Error ->
	{reply, Error, S}
    end;
handle_call({info, state}, _From,
	    S=#state{pool=#pool{workers=Ws, users=Us, num_workers=NW},
		     user_queue=UQ, num_started=NumS}) ->
  {reply, [{workers, workers_to_list(Ws)},
	   {num_workers, NW},
	   {users, Us},
	   {user_queue, user_queue_to_list(UQ)},
	   {num_started, NumS}
	  ], S};
handle_call(Request, From, State) ->
  Reply={invalid_request, Request},
  log(error, "unexpected call ~w from ~w", [Reply, From]),
  {reply, Reply, State}.
%%--------------------------------------------------------------------
%% Function: handle_cast/2
%% Description: Handling cast messages
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_cast({new_worker, NW},
	    S=#state{pool=P, user_queue=UQ, num_started=NumS}) ->
  P1 = pool_add_worker(NW, P),
  NumS1 = if (NumS > 0) -> (NumS - 1);
	     true -> NumS
	  end,
  {UQ1, P2} = assign_user(UQ, P1),
  {noreply, S#state{pool=P2, user_queue=UQ1, num_started=NumS1}};
handle_cast({disconnect, User, Worker}, S=#state{pool=P, user_queue=UQ}) ->
  P1 = pool_delete_user(Worker, User, P),
  {UQ1, P2} = assign_user(UQ, P1),
  {noreply, S#state{pool=P2, user_queue=UQ1}};
handle_cast({connect_timeout, User, Ref}, S=#state{pool=P, user_queue=UQ}) ->
  NewUQ = user_queue_delete(User, Ref, UQ),
  %may be we already replied him
  NewP = pool_delete_user(Ref, User, P),
  {noreply, S#state{pool=NewP, user_queue=NewUQ}};
handle_cast(Request, State) ->
  Info={invalid_request, Request},
  log(error, "unexpected cast ~w from ~w", [Info]),
  {noreply, State}.

%%--------------------------------------------------------------------
%% Function: handle_info/2
%% Description: Handling all non call/cast messages
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_info({'EXIT', PS, Reson}, S=#state{pool_sup=PS, pool={_, Users}}) ->
  lists:foreach(fun ({User, _, _}) -> unlink(User) end, Users),
  {stop, Reson, S};
handle_info({'EXIT', Pid, Reson},
	    S=#state{pool=P, pool_sup=PS, pool_max_len=Max, user_queue=UQ,
		     num_started=NumS}) ->
  {NUQ, NewPool} =
    case is_pool_user(Pid, P) of
      true ->
	P1 = pool_delete_user(Pid, P),
	assign_user(UQ, P1);
      false ->
	case {pool_delete_worker(Pid, P),
	      user_queue_empty(UQ),
	      ((pool_num_workers(P) + NumS) =< Max)} of
	  {{true, NP}, false, true} ->
	    start_worker(PS),
	    {UQ, NP};
	  {{_, NP}, _, _} ->
	    {UQ, NP}
	end
    end,
  {noreply, S#state{pool=NewPool, user_queue=NUQ}};
handle_info({Pid, stop}, State) when pid(Pid) ->
  {stop, normal, State};
handle_info({'DOWN', _, process, User, _}, S=#state{user_queue=UQ}) ->
  NewUQ = user_queue_delete(User, UQ),
  {noreply, S#state{pool=NewUQ}};
handle_info(Info, State) ->
  {noreply, State}.



%%--------------------------------------------------------------------
%% Function: terminate/2
%% Description: Shutdown the server
%% Returns: any (ignored by gen_server)
%%--------------------------------------------------------------------
terminate(Reason, S=#state{pool=#pool{users=Us}, pool_sup=PS,user_queue=UQ}) ->
  lists:foreach(fun ({User, _, _}) -> unlink(User) end, Us),
  lists:foreach(fun ({_, Ref}) -> erlang:demonitor(Ref) end,
		queue:to_list(UQ)),
  ok.

%%--------------------------------------------------------------------
%% Func: code_change/3
%% Purpose: Convert process state when code is changed
%% Returns: {ok, NewState}
%%--------------------------------------------------------------------
code_change(OldVsn, State, Extra) ->
  {ok, State}.

%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------

assign_user(UQ, Pool) ->
  assign_user(UQ, Pool, user_queue_empty(UQ), pool_has_free_workers(Pool)).

assign_user(UQ, Pool, false, true) ->
  {{User, ReplyTo, Ref}, UQ1} = user_queue_out(UQ),
  {{ok, W}, Pool1} = pool_select_worker(Pool),
  Pool2 = pool_add_user(W, User, Ref, Pool1),
  gen_server:reply(ReplyTo, {ok, conn_id(W, self())}),
  assign_user(UQ1, Pool2);
assign_user(UQ, Pool, _, _) ->
  {UQ, Pool}.

start_pool_sup(SupSpec, WorkerSpec) ->
  connection_pool_sup:start_link({SupSpec, WorkerSpec}).
    
init_pool(PoolSup, 0) ->
  ok;
init_pool(PoolSup, Num) ->
  start_worker(PoolSup),
  init_pool(PoolSup, Num-1).

start_worker(PoolSup) ->
  case connection_pool_sup:start_child(PoolSup) of
    {error, Error} -> % crash if server can't start
      exit(Error);
    Other ->
      Other
  end.

wrap_worker({Fun, Args}) when function(Fun) ->
  {?MODULE, worker_wrapper, [self(), Fun, Args]};
wrap_worker({Module, Fun, Args}) ->
  WrapFun = fun (As) ->
		apply(Module, Fun, Args)
	    end,
  {?MODULE, worker_wrapper, [self(), WrapFun, [Args]]}.
  
worker_wrapper(PoolMgr, Fun, Args) ->
  case apply(Fun, Args) of
    {ok, Pid} ->
      link(Pid),
      new_pool_worker(PoolMgr, Pid),
      {ok, Pid};
    Error ->
      Error
  end.

check_worker_spec({FW, Restart, Shutdown, Type, Modules}) ->
  WS = [{make_ref(), wrap_worker(FW), Restart, Shutdown, Type, Modules}],
  case supervisor:check_childspecs(WS) of
    ok ->
      {ok, WS};
    Error ->
      Error
  end;
check_worker_spec(_) ->
  {error, bad_worker_spec}.

new_pool_worker(PoolMgr, Pid) ->
  gen_server:cast(PoolMgr, {new_worker, Pid}).

%% Pool
pool_new(exclusive) ->
  #pool{workers={fun excl_pool/2, queue:new()}};
pool_new(shared) ->
  #pool{workers={fun shared_pool/2, orddict:new()}}.

%
pool_num_workers(P=#pool{num_workers=Num}) ->
  Num.

pool_has_free_workers(P=#pool{workers=Ws}) ->
  workers_have_free(Ws).

pool_select_worker(P=#pool{workers=Ws}) ->
  {Worker, NewWs} = workers_select(Ws),
  {Worker, P#pool{workers=NewWs}}.

pool_add_worker(Worker, P=#pool{workers=Ws, users=Us, num_workers=WN}) ->
  link(Worker),
  P#pool{workers=workers_add(Ws, Worker), num_workers=WN+1}.

pool_delete_worker(Worker, P=#pool{workers=Ws, users=Us, num_workers=WN}) ->
  {Del1, NewWs} = workers_delete(Ws, Worker),
  {Del2, NewUs} = lists:foldl(fun ({U, W, _}, {_, Acc}) when W == Worker ->
				  unlink(U),
				  {true, Acc};
				  (V, {D, Acc}) ->
				  {D, [V|Acc]}
			      end, {false, []}, Us),
  Deleted = (Del1 or Del2),
  NewWN = if Deleted == true -> WN - 1;
	     true -> WN
	  end,
  {Deleted, P#pool{workers=NewWs, users=NewUs, num_workers=NewWN}}.

pool_add_user(Worker, User, Ref, P=#pool{workers=Ws, users=Us}) ->
  link(User),
  NewWs = workers_add_user(Ws, Worker, User),
  NewUs = [{User, Worker, Ref}|Us],
  P#pool{workers=NewWs, users=NewUs}.

pool_delete_user(Ref, User, P) when reference(Ref) ->
  pool_delete_user(Ref, User, P, 3);
pool_delete_user(Pid, User, P) when pid(Pid) ->
  pool_delete_user(Pid, User, P, 2).

pool_delete_user(Key, Pid, P=#pool{workers=Ws, users=Us}, KeyPos) ->
  case lists:keysearch(Key, KeyPos, Us) of
    {value, U={Pid, Worker, _Ref}} ->
      unlink(Pid),
      NewWs = workers_delete_user(Ws, Worker, Pid),
      NewUs = lists:delete(U, Us),
      P#pool{workers=NewWs, users=NewUs};
    {value, U={BadPid, Worker, _Ref}} ->
      log(error, "attempt to disconnect ~w from ~w~n", [BadPid, Pid]),
      P;
    false ->
      P
  end.

pool_delete_user(User, P=#pool{workers=Ws, users=Us}) ->
  {NewWs, NewUs} =
    lists:foldl(fun ({U, W, _}, {Ws1, Us1}) when U == User ->
		    {workers_delete_user(Ws1, W, U), Us1};
		    (V, {Ws1, Us1}) ->
		    {Ws1, [V|Us1]}
		end, {Ws, []}, Us),
  P#pool{workers=NewWs, users=NewUs}.

is_pool_user(User, P=#pool{users=Us}) ->
  case lists:keysearch(User, 1, Us) of
    {value, _} ->
      true;
    false ->
      false
  end.

%%% Workers

workers_to_list({Fun, Pool}) ->
  Fun(to_list, Pool).

workers_have_free({Fun, Pool}) ->
  Fun(has_free, Pool).

workers_select({Fun, Pool}) ->
  {Ret, NewPool} = Fun(select, Pool),
  {Ret, {Fun, NewPool}}.

workers_add({Fun, Pool}, Worker) ->
  {Fun, Fun(add, {Pool, Worker})}.

workers_delete({Fun, Pool}, Worker) ->
  {Deleted, NewPool} = Fun(delete, {Pool, Worker}),
  {Deleted, {Fun, NewPool}}.

workers_add_user({Fun, Pool}, Worker, User) ->
  {Fun, Fun(add_user, {Pool, Worker, User})}.

workers_delete_user({Fun, Pool}, Worker, User) ->
  {Fun, Fun(delete_user, {Pool, Worker, User})}.

shared_pool(select, Pool) ->
  Choosed =
    orddict:fold(fun (Worker, Count, {Prev, PrevMax}) when Count < PrevMax -> 
		     {Worker, Count};
		     (Worker, Count, to_queue) ->
		     {Worker, Count};
		     (_, _, Acc) ->
		     Acc
		 end, to_queue, Pool),
  dbg:assert(Choosed == init, "~w:~w: empty pool", [?MODULE, ?LINE]),
  case Choosed of
    {Worker, _} ->
      {{ok, Worker}, Pool};
    to_queue ->
      {to_queue, Pool}
  end;
shared_pool(to_list, Pool) ->
  orddict:to_list(Pool);

shared_pool(has_free, Pool) ->
  (catch
   orddict:fold(fun (Worker, 0, Res) -> 
		    throw(true);
		    (_, _, Res) ->
		    Res
		end, false, Pool));

shared_pool(add, {Pool, Worker}) ->
  orddict:store(Worker, 0, Pool);

shared_pool(delete, {Pool, Worker}) ->
  case orddict:is_key(Worker, Pool) of
    true ->
      {true, orddict:erase(Worker, Pool)};
    false ->
      {false, Pool}
  end;

shared_pool(add_user, {Pool, Worker, _User}) ->
  orddict:update_counter(Worker, 1, Pool);

shared_pool(delete_user, {Pool, Worker, _User}) ->
  orddict:update_counter(Worker, -1, Pool).

%%%
excl_pool(select, Pool) ->
  case queue:out(Pool) of
    {empty, _} ->
      {to_queue, Pool};
    {{value, Worker}, NewPool} ->
      {{ok, Worker}, NewPool}
  end;

excl_pool(has_free, Pool) ->
  case queue:out(Pool) of
    {empty, _} ->
      false;
    _ ->
      true
  end;

excl_pool(to_list, Pool) ->
  [{Pid, 0} || Pid <- queue:to_list(Pool)];

excl_pool(add, {Pool, Worker}) ->
  queue:in(Worker, Pool);

excl_pool(delete, {Pool, Worker}) ->
  {Deleted, NewPool} =
    lists:foldl(fun (W, {D, Acc}) when W == Worker ->
		    {true, Acc};
		    (V, {D, Acc}) ->
		    {D, queue:in(V, Acc)}
		end, {false, queue:new()}, queue:to_list(Pool));

excl_pool(add_user, {Pool, _Worker, _User}) ->
  Pool;

excl_pool(delete_user, {Pool, Worker, _User}) ->
  queue:in(Worker, Pool).

%%%
user_queue_empty(Q) ->
  case queue:out(Q) of
    {empty, _} ->
      true;
    _ ->
      false
  end.

user_queue_in(User={Pid, _, _}, Q) ->
  Ref = erlang:monitor(process, Pid),
  queue:in({User, Ref}, Q).

user_queue_out(Q) ->
  case queue:out(Q) of
    {{value, {User, Ref}}, Q1} ->
      erlang:demonitor(Ref),
      {User, Q1};
    Result -> %{empty, Q1}
      Result
  end.

user_queue_delete(User, Q) ->
  NewQ =
    lists:foldl(fun ({{U, _, _}, Ref}, Acc) when U == User ->
		    erlang:demonitor(Ref),
		    Acc;
		    (V, Acc) ->
		    queue:in(V, Acc)
		end, queue:new(), queue:to_list(Q)),
  NewQ.

user_queue_delete(User, Ref, Q) ->
  NewQ =
    lists:foldl(fun ({{U, _, R}, MRef}, Acc) when U == User, R == Ref ->
		    erlang:demonitor(MRef),
		    Acc;
		    (V, Acc) ->
		    queue:in(V, Acc)
		end, queue:new(), queue:to_list(Q)),
  NewQ.

user_queue_to_list(Q) ->
  queue:to_list(Q).

conn_id(Worker, Mgr) ->
  #conn_id{pool_pid=Mgr, worker_pid=Worker}.

log(error, Fmt, Args) ->  
  error_logger:format(Fmt, Args).

options(Opts) ->
  MkChkLen = fun (Min) ->
		 fun (V) when integer(V), V >= Min ->
		     {ok, V};
		     (_) ->
		     {error, bad_uinteger}
		 end
	     end,
  ChkType = fun (Type) when Type == shared; Type == exclusive ->
		{ok, Type};
		(Type) ->
		{error, {bad_pool_type, Type}}
	    end,
  getopt:options(Opts,
		 [{worker_spec, required, fun check_worker_spec/1},
		  {supervisor_spec, {simple_one_for_one, 1,  1000}, term},
		  {pool_type, exclusive, ChkType},
		  {pool_max_len, 1, MkChkLen(1)},
		  {pool_start_len, 0, MkChkLen(0)}],
		 defined_strict).
