%%%-------------------------------------------------------------------
%%% File    : mnesia_dbi.erl
%%% Author  :  <svg@surnet.ru>
%%% Description : Misc Mnesia helper funs
%%%
%%% Created :  5 Jan 2003 by  <svg@surnet.ru>
%%%-------------------------------------------------------------------
%%
%% $Id$
%%
%% $Log$
%%
%%**
%% 
%% @doc
%% @end
%%*
-module(svglib.mnesia_dbi).

-export([schema_create/2, table_delete/3, table_add/4]).
-export([init_sequences/1, create_seq/1, create_seq/2, seq_next/1, seq_next/2, seq_current/1]).
-export([create_constraints/1]).

-import(.lists).
-import(.mnesia).
-import(.rdbms).
-import(.error_logger).
-import(.rpc).

-define(table_def(TblRec, TblProp),
	{TblRec, [{attributes, record_info(fields, TblRec)}|TblProp]}).

%% Sequences for tables
-record(table_sequence, {table,
			 val
			}).

%% Create schema and tables on Nodes::[atom()] from table definitions
%% Defs::[table_def()]
%% All tables are created as disk copies
%% It throws error if something is wrong
schema_create(Nodes, Defs) when list(Nodes) ->
  ok = mnesia:create_schema(Nodes),
  mnesia:start(),
  rpc:multicall(Nodes -- [node()], mnesia, start, []),
  CreateTable =
    fun ({Name, Attrs}) ->
        case mnesia:create_table(Name, [{disc_copies, Nodes}|Attrs]) of
          {atomic, _} -> ok;
          Error -> throw(error(schema_create, Error))
        end
    end,
  lists:foreach(CreateTable, Defs),
  ok.

table_delete(Nodes, TableNames, WaitTimeout) ->
  {_, Down} = rpc:multicall(Nodes, mnesia, start, []),

  wait_for_tables(TableNames, WaitTimeout),

  Up = Nodes -- Down,

  DeleteTableCopy =
    fun (Node, TableName) ->
        case mnesia:del_table_copy(TableName, Node) of
          {aborted, {badarg, _,unknown}} -> ok;
          {atomic, _} -> ok;
          Error -> throw(error(schema_delete,Error))
        end
    end,
  listn:mapn(DeleteTableCopy, [Up, TableNames]),
  Up.

%% if node hasn't schema it creates one.
table_add(MasterNode, Nodes, TableNames, WaitTimeout) ->

  {_, Down} = rpc:multicall(Nodes, mnesia, start, []),

  wait_for_tables(TableNames, WaitTimeout),

  Up = Nodes -- Down,
  NoSchemaNodes = Up -- mnesia:system_info(running_db_nodes),

  MakeUsKnown = 
    fun(Node) ->
        case rpc:call(Node, mnesia, change_config,
                      [extra_db_nodes, [MasterNode]]) of
          {badrpc, Reson} ->
            throw({error, {Node, Reson}});
          {ok, [MasterNode]} ->
            ok;
          Error ->
            throw(Error)
        end
    end,

  AddTableCopy =
    fun (Node, TableName) ->
        case mnesia:add_table_copy(TableName, Node, disc_copies) of
          {aborted, {already_exists, _, _}} -> ok;
          {atomic, _} -> ok;
          Error -> throw(Error)
        end
    end,

  ReplOurSchema =
    fun(Node) ->
        ok = MakeUsKnown(Node),
        case not lists:member(Node,
                              mnesia:table_info(schema, disc_copies)) of
          true ->
            {atomic, ok} =
              mnesia:change_table_copy_type(schema, Node, disc_copies);
          _ ->
            ok
        end
    end,
  lists:foreach(ReplOurSchema, NoSchemaNodes),
  listn:mapn(AddTableCopy, [Up, TableNames]),
  {_, []} = rpc:multicall(Up, mnesia, set_master_nodes, [MasterNode]),
  Up.

wait_for_tables(Tables, Timeout) ->
  ForceLoad = fun(Table, Bad) ->
                  case mnesia:force_load_table(Table) of
                    yes ->
                      log_info("Force load table ~p", [Table]),
                      Bad;
                    Error ->
                      [{Table, Error}|Bad]
                  end
              end,
  case mnesia:wait_for_tables(Tables,Timeout) of
    {timeout,BadTables} ->
      case lists:foldl(ForceLoad, [], BadTables) of
        [] ->
          ok;
        FatalTables ->
          error({force_load, [element(1, T) || T <- FatalTables]},
                {force_load, FatalTables})
      end;
    {error,Reason} ->
      error(not_loaded, Reason);
    ok ->
      ok
  end.

create_constraints(Cs) ->
  create_constraints(Cs, []).

create_constraints([], Props) ->
  rdbms:add_properties(Props);
create_constraints([{Class, Name, Attrs}|Rest], Props) ->
  create_constraints(Rest, 
		     [{{Class, Name, Prop}, Val} || {Prop, Val} <- Attrs]
		     ++ Props).

init_sequences(Nodes) ->
  {Tbl, Attrs} = ?table_def(table_sequence, [{type, set}]),
  case mnesia:create_table(Tbl, [{disc_copies, Nodes}|Attrs]) of
    {atomic, _} -> ok;
    Error -> exit(Error)
  end.

create_seq(Tbl) ->
  create_seq(Tbl, 0).

create_seq(Tbl, InitVal) when integer(InitVal) ->
  Tr = fun () ->
	   mnesia:write(#table_sequence{table=Tbl, val=InitVal})
       end,
  case mnesia:transaction(Tr) of
    {atomic, _} ->
      ok;
    {aborted, Error} ->
      exit(Error)
  end.

seq_next(Tbl) ->
  seq_next(Tbl, 1).

seq_next(Tbl, Val) when integer(Val) ->
  mnesia:dirty_update_counter({table_sequence, Tbl}, Val).

seq_current(Tbl) ->
  [#table_sequence{val=Ret}]=mnesia:dirty_read({table_sequence, Tbl}),
  Ret.

error(Info, Desc) ->
  {error, {Info, Desc}}.

log_info(Format, Args) ->
  case application:get_application() of
    {ok, App} ->
      error_logger:info_msg("~s, ~s:" ++ Format, [App, ?MODULE|Args]);
    _ ->
      error_logger:info_msg("~s:" ++ Format, [?MODULE|Args])
  end.
